from __future__ import annotations

import collections
import functools
import os
from typing import Any, Dict, Iterable, Optional, Set

import datasets
import meds
import polars as pl

import femr.hf_utils


def _get_all_codes_map(batch) -> Set[str]:
    result = set()
    for events in batch["events"]:
        for event in events:
            for measurement in event["measurements"]:
                result.add(measurement["code"])
    return result


def _get_all_codes_agg(first: Set[str], second: Set[str]) -> Set[str]:
    first |= second
    return first


class Ontology:
    def __init__(self, athena_path: str, code_metadata: meds.CodeMetadata = {}):
        """Create an Ontology from an Athena download and an optional meds Code Metadata structure.

        NOTE: This is an expensive operation.
        It is recommended to create an ontology once and then save/load it as necessary.
        """
        # Load from code metadata
        self.description_map: Dict[str, str] = {}
        self.parents_map: Dict[str, Set[str]] = collections.defaultdict(set)

        # Load from the athena path ...
        concept = pl.scan_csv(os.path.join(athena_path, "CONCEPT.csv"), separator="\t", infer_schema_length=0)
        code_col = pl.col("vocabulary_id") + "/" + pl.col("concept_code")
        description_col = pl.col("concept_name")
        concept_id_col = pl.col("concept_id").cast(pl.Int64)

        processed_concepts = (
            concept.select(code_col, concept_id_col, description_col, pl.col("standard_concept").is_null())
            .collect()
            .rows()
        )

        concept_id_to_code_map = {}

        non_standard_concepts = set()

        for code, concept_id, description, is_non_standard in processed_concepts:
            concept_id_to_code_map[concept_id] = code

            # We don't want to override code metadata
            if code not in self.description_map:
                self.description_map[code] = description

            if is_non_standard:
                non_standard_concepts.add(concept_id)

        relationship = pl.scan_csv(
            os.path.join(athena_path, "CONCEPT_RELATIONSHIP.csv"), separator="\t", infer_schema_length=0
        )
        relationship_id = pl.col("relationship_id")
        relationship = relationship.filter(
            relationship_id == "Maps to", pl.col("concept_id_1") != pl.col("concept_id_2")
        )
        for concept_id_1, concept_id_2 in (
            relationship.select(pl.col("concept_id_1").cast(pl.Int64), pl.col("concept_id_2").cast(pl.Int64))
            .collect()
            .rows()
        ):
            if concept_id_1 in non_standard_concepts:
                self.parents_map[concept_id_to_code_map[concept_id_1]].add(concept_id_to_code_map[concept_id_2])

        ancestor = pl.scan_csv(os.path.join(athena_path, "CONCEPT_ANCESTOR.csv"), separator="\t", infer_schema_length=0)
        ancestor = ancestor.filter(pl.col("min_levels_of_separation") == "1")
        for concept_id, parent_concept_id in (
            ancestor.select(
                pl.col("descendant_concept_id").cast(pl.Int64), pl.col("ancestor_concept_id").cast(pl.Int64)
            )
            .collect()
            .rows()
        ):
            self.parents_map[concept_id_to_code_map[concept_id]].add(concept_id_to_code_map[parent_concept_id])

        # Have to add after OMOP to overwrite ...
        for code, code_info in code_metadata.items():
            if code_info.get("description") is not None:
                self.description_map[code] = code_info["description"]
            if code_info.get("parent_codes") is not None:
                self.parents_map[code] = set(code_info["parent_codes"])

        self.children_map = collections.defaultdict(set)
        for code, parents in self.parents_map.items():
            for parent in parents:
                self.children_map[parent].add(code)

        self.all_parents_map: Dict[str, Set[str]] = {}
        self.all_children_map: Dict[str, Set[str]] = {}

    def prune_to_dataset(
        self,
        dataset: datasets.Dataset,
        num_proc: int = 1,
        prune_all_descriptions: bool = False,
        remove_ontologies: Set[str] = set(),
    ) -> None:
        valid_codes = femr.hf_utils.aggregate_over_dataset(
            dataset,
            functools.partial(_get_all_codes_map),
            _get_all_codes_agg,
            num_proc=num_proc,
            batch_size=1_000,
        )

        if prune_all_descriptions:
            self.description_map = {}

        all_parents = set()

        for code in valid_codes:
            all_parents |= self.get_all_parents(code)

        def is_valid(code):
            ontology = code.split("/")[0]
            return (code in valid_codes) or ((ontology not in remove_ontologies) and (code in all_parents))

        codes = self.children_map.keys() | self.parents_map.keys() | self.description_map.keys()
        for code in codes:
            m: Any
            if is_valid(code):
                for m in (self.children_map, self.parents_map):
                    m[code] = {a for a in m[code] if is_valid(a)}
            else:
                for m in (self.children_map, self.parents_map, self.description_map):
                    if code in m:
                        del m[code]

        self.all_parents_map = {}
        self.all_children_map = {}

        # Prime the pump
        for code in self.children_map.keys() | self.parents_map.keys():
            self.get_all_parents(code)

    def get_description(self, code: str) -> Optional[str]:
        """Get a description of a code."""
        return self.description_map.get(code)

    def get_children(self, code: str) -> Iterable[str]:
        """Get the children for a given code."""
        return self.children_map.get(code, set())

    def get_parents(self, code: str) -> Iterable[str]:
        """Get the parents for a given code."""
        return self.parents_map.get(code, set())

    def get_all_children(self, code: str) -> Set[str]:
        """Get all children, including through the ontology."""
        if code not in self.all_children_map:
            result = {code}
            for child in self.children_map.get(code, set()):
                result |= self.get_all_children(child)
            self.all_children_map[code] = result
        return self.all_children_map[code]

    def get_all_parents(self, code: str) -> Set[str]:
        """Get all parents, including through the ontology."""
        if code not in self.all_parents_map:
            result = {code}
            for parent in self.parents_map.get(code, set()):
                result |= self.get_all_parents(parent)
            self.all_parents_map[code] = result

        return self.all_parents_map[code]



from __future__ import annotations

import collections
import datetime
import functools
from typing import Any, Dict, List, Mapping, Tuple, Optional

import datasets
import numpy as np
import torch.utils.data

import meds

import femr.hf_utils
import femr.models.tokenizer
import femr.pat_utils


def map_preliminary_batch_stats(batch, indices, *, processor: FEMRBatchProcessor, max_length: int):
    """
    This function creates preliminary batch statistics, to be used for final batching.

    The overall problem this is trying to solve is that Patient records can be very long, so when we featurize patients we actually featurize subsequences of patients.

    AKA every patient in a batch is actually (patient_id, start_index, length), which specifies a subsequence of the patient.

    The trickiness becomes when a particular patient has multiple labels, some of which will require multiple subsequences.

    The goal of this function is to compute the list [(patient_id, start_index, length)] such that every label is covered
    by at least one batch. Note that some labels will be covered multiple times.

    Note that there is a special setting for tasks that don't need exact labeling (needs_exact in tasks.py returns False).
    For these patients we only generate one tuple for them, and drop some labels.

    Later code will then take [(patient_id, start_index, length)], and create actual batches.
    """
    lengths = []

    for patient_index, patient_id, events in zip(indices, batch["patient_id"], batch["events"]):
        patient = {
            "patient_id": patient_id,
            "events": events,
        }
        data = processor.convert_patient(patient)

        # There are no labels for this patient
        if data["transformer"]["label_indices"].shape[0] == 0:
            continue

        # We need exact batching, so we need an algorithm that precisely covers every label in the batch
        if data["needs_exact"]:
            current_start = 0
            current_end = 0
            for label_index in data["transformer"]["label_indices"]:
                if (label_index - current_start + 1) >= max_length:
                    if current_start != current_end:
                        lengths.append((patient_index, current_start, current_end - current_start + 1))
                    current_start = label_index - max_length + 1
                    current_end = label_index
                else:
                    current_end = label_index

            lengths.append((patient_index, current_start, current_end - current_start + 1))
        else:
            last_index = data["transformer"]["label_indices"][-1]
            length = min(max_length, last_index + 1)
            lengths.append((patient_index, last_index + 1 - length, length))
    if len(lengths) > 0:
        return [np.array(lengths, dtype=np.int64)]
    else:
        return []


def agg_preliminary_batch_stats(lengths1, lengths2):
    """Aggregate preliminary length statistics from the map_preliminary_batch_stats"""
    lengths1.extend(lengths2)

    return lengths1


class BatchCreator:
    """The BatchCreator is designed to generate batches from patient data."""
    def __init__(self, tokenizer: femr.models.FEMRTokenizer, task: Optional[femr.tasks.Task] = None):
        """Initialize a BatchCreator, with a tokenizer, and optionally a task."""
        self.tokenizer = tokenizer
        self.task = task

    def start_batch(self):
        """Start a batch."""
        self.patient_ids = []
        self.offsets = []
        self.patient_lengths = []

        if not self.tokenizer.is_hierarchical:
            self.tokens = []
        else:
            self.hierarchical_tokens = []
            self.hierarchical_weights = []
            self.token_indices = [0]

        self.valid_tokens = []

        self.ages = []
        self.normalized_ages = []
        self.timestamps = []

        self.label_indices = []

        if self.task is not None:
            self.task.start_batch()

    def add_patient(self, patient: meds.Patient, offset: int = 0, max_length: Optional[int] = None):
        """Add a patient to the current batch.
        
        Note that the two optional parameters are used to add a subset of a patient to a batch.
        It is generally recommended to never manually use offset or max_length as you should rely on FEMRBatchProcessor.convert_dataset.

        Arguments:
            patient: The patient to add.
            offset: The offset into the patient to featurize.
            max_length: The maximum length of the batch sequence. There is no max when left at None. 

        """
        current_date = None
        last_time = None


        # The overall algorithm here is a bit complex
        # First we featurize the entire patient
        # Then we slice the patient indices according to offset and max_length

        # These are the indices of the labels into the patient vectors
        per_patient_label_indices = []

        # The ages at each index for the patient
        per_patient_ages = []

        # The normalized age at index for the patient
        per_patient_normalized_ages = []

        # The timestamps at each index for the patient
        per_patient_timestamps = []

        # For a regular tokenizer, we just have tokens
        per_patient_tokens = []

        # For a hierarchical tokenizer, we have a more complex setup
        # These are designed to match the inputs required for an EmbeddingBag.
        # See PyTorch's EmbeddingBag documentation to understand what these mean.
        per_patient_hierarchical_tokens = []
        per_patient_hierarchical_weights = []
        per_patient_token_indices = [0]

        if self.task is not None:
            self.task.start_patient(patient, self.tokenizer.ontology)

        birth = femr.pat_utils.get_patient_birthdate(patient)
        self.tokenizer.start_patient()

        for event in patient["events"]:
            # We want to avoid duplicate codes in the same day, so we maintain codes_seen_today
            if event["time"].date() != current_date:
                current_date = event["time"].date()
                codes_seen_today = set()

            for measurement in event["measurements"]:
                # Get features and weights for the current event
                features, weights = self.tokenizer.get_feature_codes(event["time"], measurement)

                # Ignore events with no features
                if len(features) == 0:
                    continue

                # Ignore events where all features have already occurred
                if all(feature in codes_seen_today for feature in features):
                    continue

                codes_seen_today |= set(features)

                if (self.task is not None) and (last_time is not None):
                    # Now we have to consider whether or not to have labels for this time step
                    # The add_event function returns how many labels to assign for this time
                    num_added = self.task.add_event(last_time, event["time"], features)
                    for _ in range(num_added):
                        per_patient_label_indices.append(len(per_patient_ages) - 1)

                if not self.tokenizer.is_hierarchical:
                    assert len(features) == 1
                    per_patient_tokens.append(features[0])
                else:
                    per_patient_hierarchical_tokens.extend(features)
                    per_patient_hierarchical_weights.extend(weights)
                    per_patient_token_indices.append(len(per_patient_hierarchical_tokens))

                per_patient_ages.append((event["time"] - birth) / datetime.timedelta(days=1))
                per_patient_normalized_ages.append(self.tokenizer.normalize_age(event["time"] - birth))
                per_patient_timestamps.append(event["time"].replace(tzinfo=datetime.timezone.utc).timestamp())

                last_time = event["time"]

                
        if self.task is not None and last_time is not None:
            num_added = self.task.add_event(last_time, None, None)
            for _ in range(num_added):
                per_patient_label_indices.append(len(per_patient_ages) - 1)

        # Now we want to actually add the patient data to the batch.
        # This will involve some clever slicing.

        # First, let's get the length we are adding
        length_found = len(per_patient_ages)
        if max_length is not None:
            length_to_add = min(length_found - offset, max_length)
        else:
            length_to_add = length_found - offset

        start_index = len(self.ages)

        # Let's add the constants first
        self.valid_tokens.extend([True] * length_to_add)
        self.patient_ids.extend([patient['patient_id']] * length_to_add)
        self.offsets.append(offset)
        self.patient_lengths.append(length_to_add)

        # Ages, normalized ages and timestamps are also easy to add
        self.ages.extend(per_patient_ages[offset: offset + length_to_add])
        self.normalized_ages.extend(per_patient_normalized_ages[offset: offset + length_to_add])
        self.timestamps.extend(per_patient_timestamps[offset: offset + length_to_add])

        if not self.tokenizer.is_hierarchical:
            # Easy for simple tokenizer
            self.tokens.extend(per_patient_tokens[offset: offset + length_to_add])
        else:
            # Hierarchical tokenizer is more complex since we have to shift the indices as well
            # Remember, these arrays are all designed for PyTorch EmbeddingBag

            # We need to get the start and end at a particular offset
            internal_start = per_patient_token_indices[offset]
            internal_end = per_patient_token_indices[offset + length_to_add]
            
            # We need to offset the token indices to account for the existing tokens            
            self.token_indices.extend([len(self.hierarchical_tokens) - internal_start + value for value in per_patient_token_indices[offset + 1:offset + length_to_add + 1]])

            self.hierarchical_tokens.extend(per_patient_hierarchical_tokens[internal_start: internal_end])
            self.hierarchical_weights.extend(per_patient_hierarchical_weights[internal_start: internal_end])


        # The label indices are also a bit tricky as they have to be offset accordingly.
        # We also need to collect good labels that should be sent to the final numpy arrays.
        labels_to_add = []
        for i, label_index in enumerate(per_patient_label_indices):
            corrected_label = label_index - offset

            if 0 <= corrected_label < length_to_add:
                labels_to_add.append(i)
                self.label_indices.append(start_index + corrected_label)

        if self.task is not None:
            self.task.add_patient_labels(labels_to_add)


    def get_batch_data(self):
        """Convert the batch to numpy arrays. The data structure is defined inline in this function."""
        if self.tokenizer.vocab_size <= 2**15:
            token_dtype = np.int16
        else:
            token_dtype = np.int32

        transformer = {
            # Whether or not the token is valid at this index
            "valid_tokens": np.array(self.valid_tokens),

            # The age of the patient in days at this index
            "ages": np.array(self.ages, dtype=np.float32),

            # The normalized ages at this index
            "normalized_ages": np.array(self.normalized_ages, dtype=np.float16),

            # The timestamp (in seconds) at this index
            "timestamps": np.array(self.timestamps, dtype=np.int64),

            # The length of the patient
            "patient_lengths": np.array(self.patient_lengths, dtype=np.int32),

            # The indices of the labels
            "label_indices": np.array(self.label_indices, dtype=np.int32),
        }

        if not self.tokenizer.is_hierarchical:
            # For a single tokenizer, these are simple the token indices
            transformer["tokens"] = np.array(self.tokens, dtype=token_dtype)
        else:
            # See PyTorch's EmbeddingBag for what these numpy arrays mean.
            transformer["hierarchical_tokens"] = np.array(self.hierarchical_tokens, dtype=token_dtype)
            transformer["hierarchical_weights"] = np.array(self.hierarchical_weights, dtype=np.float16)
            transformer["token_indices"] = np.array(self.token_indices, dtype=np.int32)

        # Some general metadata
        final = {
            "num_patients": len(self.patient_lengths),
            "num_indices": len(self.label_indices),
            "patient_ids": np.array(self.patient_ids, dtype=np.int64),
            "offsets": np.array(self.offsets, dtype=np.int32),
            "transformer": transformer,
        }

        # Add the task data
        if self.task is not None and transformer["label_indices"].shape[0] > 0:
            final["task"] = self.task.get_batch_data()
            final["needs_exact"] = self.task.needs_exact()
        return final

    def cleanup_batch(self, batch: Dict[str, Any]) -> Dict[str, Any]:
        """Clean a batch, applying final processing.

        This is necessary as some tasks use sparse matrices that need to be postprocessed."""

        batch["transformer"]["patient_lengths"] = np.array(batch["transformer"]["patient_lengths"])
        assert isinstance(batch["transformer"]["patient_lengths"], np.ndarray)

        if self.task is not None and "task" in batch:
            batch["task"] = self.task.cleanup(batch["task"])

        return batch


def _batch_generator(batch_data: Tuple[np.ndarray, np.ndarray], *, creator: BatchCreator, dataset: datasets.Dataset):
    for lengths, offsets in batch_data:
        offsets = list(offsets)
        for start, end in zip(offsets, offsets[1:]):
            creator.start_batch()
            for patient_index, offset, length in lengths[start:end, :]:
                creator.add_patient(dataset[patient_index.item()], offset, length)

            result = creator.get_batch_data()
            assert "task" in result, f"No task present in {lengths[start:end,:]}"

            yield result


def _add_dimension(data: Any) -> Any:
    if isinstance(data, collections.abc.Mapping):
        return {k: _add_dimension(v) for k, v in data.items()}
    elif isinstance(data, torch.Tensor):
        return data.unsqueeze(dim=0)
    elif isinstance(data, np.ndarray):
        return np.expand_dims(data, axis=0)
    elif isinstance(data, (int, float, np.number, np.bool_)):
        return data
    else:
        raise RuntimeError("Could not convert item of type " + str(type(data)))


class FEMRBatchProcessor:
    """The FEMR Batch processor creates batches for processing by a transformer."""
    def __init__(self, tokenizer, task=None):
        self.creator = BatchCreator(tokenizer, task)

    def convert_patient(self, patient: meds.Patient, offset: int =0, max_length: Optional[int]=None, tensor_type=None, **formatter_kwargs):
        """Convert a single patient to a batch.
        
        Note that this can also convert parts of a patient to a batch using the offset and max_length parameters.
        This is useful for processing long patients.

        NOTE: This function is primarily for debugging purposes. It is recommended to use convert_dataset for maximum correctness and efficiency.

        Arguments:
            patient: The patient to convert
            offset: The integer offset into the patient to convert
            max_length: The maximum length to convert
            tensor_type: The dataset to return
            formatter_kwargs: Arguments for a datasets formatter when converting datatypes
            
        Returns:
            A batch, ready to be fed into a FEMR transformer model
        """
        self.creator.start_batch()
        self.creator.add_patient(patient, offset=offset, max_length=max_length)
        batch_data = self.creator.get_batch_data()
        if tensor_type is not None:
            formatter = datasets.formatting.get_formatter(tensor_type, **formatter_kwargs)
            batch_data = formatter.recursive_tensorize(batch_data)
        return batch_data

    def collate(self, batches: List[Mapping[str, Any]]) -> Mapping[str, Any]:
        """A collate function that prepares batches for being fed into a dataloader."""
        assert len(batches) == 1, "Can only have one batch when collating"
        return {"batch": _add_dimension(self.creator.cleanup_batch(batches[0]))}

    def convert_dataset(self, dataset, tokens_per_batch: int, min_patients_per_batch: int = 4, num_proc: int = 1):
        """Convert an entire dataset to batches.
        
        Arguments:
            dataset: A huggingface dataset containing MEDS patients
            tokens_per_batch: The number of tokens allowed per batch
            min_patients_per_batch: The minimum number of patients per batch
            num_proc: The number of processers to use when converting
        
        Returns:
            A huggingface dataset object containing batches
        """
        if isinstance(dataset, datasets.DatasetDict):
            return datasets.DatasetDict(
                {
                    k: self.convert_dataset(v, tokens_per_batch, min_patients_per_batch, num_proc)
                    for k, v in dataset.items()
                }
            )

        max_length = tokens_per_batch // min_patients_per_batch
        lengths = femr.hf_utils.aggregate_over_dataset(
            dataset,
            functools.partial(map_preliminary_batch_stats, processor=self, max_length=max_length),
            agg_preliminary_batch_stats,
            num_proc=num_proc,
            batch_size=1_000,
            with_indices=True,
        )

        lengths = np.concatenate(lengths)

        rng = np.random.default_rng()
        rng.shuffle(lengths)

        current_batch_length = 0

        batch_offsets = [0]

        for i, length in enumerate(lengths[:, 2]):
            if current_batch_length + length > tokens_per_batch:
                batch_offsets.append(i)
                current_batch_length = 0

            current_batch_length += length

        batch_offsets.append(len(lengths))

        batches = list(zip(batch_offsets, batch_offsets[1:]))

        split_batches = np.array_split(batches, num_proc)

        final_batch_data = []

        for batch_part in split_batches:
            if len(batch_part) == 0:
                continue
            start = batch_part[0][0]
            end = batch_part[-1][-1]
            lengths_part = lengths[start:end, :]
            offsets = [0] + [b - start for _, b in batch_part]

            final_batch_data.append(
                (
                    lengths_part,
                    np.array(offsets, dtype=np.int32),
                )
            )

        print("Creating batches", len(batches))

        batch_func = functools.partial(
            _batch_generator,
            creator=self.creator,
            dataset=dataset,
        )

        batch_dataset = datasets.Dataset.from_generator(
            batch_func,
            gen_kwargs={
                "batch_data": final_batch_data,
            },
            num_proc=num_proc,
            writer_batch_size=8,
        )

        return batch_dataset



# flake8: noqa: E402

import datetime
import os
import pathlib
import sys
import warnings
from typing import List

import meds
from femr_test_tools import EventsWithLabels, run_test_for_labeler

from femr.labelers import TimeHorizon, TimeHorizonEventLabeler


class DummyLabeler(TimeHorizonEventLabeler):
    """Dummy labeler that returns True if the event's `code` is in `self.outcome_codes`."""

    def __init__(self, outcome_codes: List[int], time_horizon: TimeHorizon, allow_same_time: bool = True):
        self.outcome_codes: List[str] = [str(a) for a in outcome_codes]
        self.time_horizon: TimeHorizon = time_horizon
        self.allow_same_time = allow_same_time

    def allow_same_time_labels(self) -> bool:
        return self.allow_same_time

    def get_prediction_times(self, patient: meds.Patient) -> List[datetime.datetime]:
        return sorted(list({e["time"] for e in patient["events"]}))

    def get_time_horizon(self) -> TimeHorizon:
        return self.time_horizon

    def get_outcome_times(self, patient: meds.Patient) -> List[datetime.datetime]:
        times: List[datetime.datetime] = []
        for e in patient["events"]:
            for m in e["measurements"]:
                if m["code"] in self.outcome_codes:
                    times.append(e["time"])
        return times


def test_no_outcomes(tmp_path: pathlib.Path):
    # No outcomes occur in this patient's timeline
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))
    labeler = DummyLabeler([100], time_horizon)
    events_with_labels: EventsWithLabels = [
        # fmt: off
        (((2015, 1, 3), 2, None), "duplicate"),
        (((2015, 1, 3), 1, None), "duplicate"),
        (((2015, 1, 3), 3, None), False),
        (((2015, 10, 5), 1, None), False),
        (((2018, 1, 3), 2, None), False),
        (((2018, 3, 3), 1, None), False),
        (((2018, 5, 3), 2, None), False),
        (((2018, 5, 3, 11), 1, None), False),
        (((2018, 5, 4), 1, None), False),
        (((2018, 12, 4), 1, None), "out of range"),
        # fmt: on
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_no_outcomes")


def test_horizon_0_180_days(tmp_path: pathlib.Path):
    # (0, 180) days
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))
    labeler = DummyLabeler([2], time_horizon)
    events_with_labels: EventsWithLabels = [
        # fmt: off
        (((2015, 1, 3), 2, None), "duplicate"),
        (((2015, 1, 3), 1, None), "duplicate"),
        (((2015, 1, 3), 3, None), True),
        (((2015, 10, 5), 1, None), False),
        (((2018, 1, 3), 2, None), True),
        (((2018, 3, 3), 1, None), True),
        (((2018, 5, 3), 2, None), True),
        (((2018, 5, 3, 11), 1, None), False),
        (((2018, 5, 4), 1, None), False),
        (((2018, 12, 4), 1, None), "out of range"),
        # fmt: on
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_horizon_0_180_days")


def test_horizon_0_180_days_no_same(tmp_path: pathlib.Path):
    # (0, 180) days
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))
    labeler = DummyLabeler([2], time_horizon, allow_same_time=False)
    events_with_labels: EventsWithLabels = [
        # fmt: off
        (((2015, 1, 3), 2, None), "duplicate"),
        (((2015, 1, 3), 1, None), "duplicate"),
        (((2015, 1, 3), 3, None), "same"),
        (((2015, 10, 5), 1, None), False),
        (((2018, 1, 3), 2, None), "same"),
        (((2018, 3, 3), 1, None), True),
        (((2018, 5, 3), 2, None), "same"),
        (((2018, 5, 3, 11), 1, None), False),
        (((2018, 5, 4), 1, None), False),
        (((2018, 12, 4), 1, None), "out of range"),
        # fmt: on
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_horizon_0_180_days")


def test_horizon_1_180_days(tmp_path: pathlib.Path):
    # (1, 180) days
    time_horizon = TimeHorizon(datetime.timedelta(days=1), datetime.timedelta(days=180))
    labeler = DummyLabeler([2], time_horizon)
    events_with_labels: EventsWithLabels = [
        # fmt: off
        (((2015, 1, 3), 2, None), "duplicate"),
        (((2015, 1, 3), 1, None), "duplicate"),
        (((2015, 1, 3), 3, None), False),
        (((2015, 10, 5), 1, None), False),
        (((2018, 1, 3), 2, None), True),
        (((2018, 3, 3), 1, None), True),
        (((2018, 5, 3), 2, None), False),
        (((2018, 5, 3, 11), 1, None), False),
        (((2018, 5, 4), 1, None), False),
        (((2018, 12, 4), 1, None), "out of range"),
        # fmt: on
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_horizon_1_180_days")


def test_horizon_180_365_days(tmp_path: pathlib.Path):
    # (180, 365) days
    time_horizon = TimeHorizon(datetime.timedelta(days=180), datetime.timedelta(days=365))
    labeler = DummyLabeler([2], time_horizon)
    events_with_labels: EventsWithLabels = [
        # fmt: off
        (((2000, 1, 3), 2, None), True),
        (((2000, 10, 5), 2, None), False),
        (((2002, 1, 5), 2, None), True),
        (((2002, 3, 1), 1, None), True),
        (((2002, 4, 5), 3, None), True),
        (((2002, 4, 12), 1, None), True),
        (((2002, 12, 5), 2, None), False),
        (((2002, 12, 10), 1, None), False),
        (((2004, 1, 10), 2, None), False),
        (((2008, 1, 10), 2, None), "out of range"),
        # fmt: on
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_horizon_180_365_days")


def test_horizon_0_0_days(tmp_path: pathlib.Path):
    # (0, 0) days
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=0))
    labeler = DummyLabeler([2], time_horizon)
    events_with_labels: EventsWithLabels = [
        # fmt: off
        (((2015, 1, 3), 2, None), "duplicate"),
        (((2015, 1, 3), 1, None), True),
        (((2015, 1, 4), 1, None), False),
        (((2015, 1, 5), 2, None), True),
        (((2015, 1, 5, 10), 1, None), False),
        (((2015, 1, 6), 2, None), True),
        # fmt: on
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_horizon_0_0_days")


def test_horizon_10_10_days(tmp_path: pathlib.Path):
    # (10, 10) days
    time_horizon = TimeHorizon(datetime.timedelta(days=10), datetime.timedelta(days=10))
    labeler = DummyLabeler([2], time_horizon)
    events_with_labels: EventsWithLabels = [
        # fmt: off
        (((2015, 1, 3), 2, None), False),
        (((2015, 1, 13), 1, None), True),
        (((2015, 1, 23), 2, None), True),
        (((2015, 2, 2), 2, None), False),
        (((2015, 3, 10), 1, None), True),
        (((2015, 3, 20), 2, None), False),
        (((2015, 3, 29), 2, None), "out of range"),
        (((2015, 3, 30), 1, None), "out of range"),
        # fmt: on
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_horizon_10_10_days")


def test_horizon_0_1000000_days(tmp_path: pathlib.Path):
    # (0, 1000000) days
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=1000000))
    labeler = DummyLabeler([2], time_horizon)
    events_with_labels: EventsWithLabels = [
        # fmt: off
        (((2000, 1, 3), 2, None), True),
        (((2001, 10, 5), 1, None), True),
        (((2020, 10, 5), 2, None), True),
        (((2021, 10, 5), 1, None), True),
        (((2050, 1, 10), 2, None), True),
        (((2051, 1, 10), 1, None), False),
        (((5000, 1, 10), 1, None), "out of range"),
        # fmt: on
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_horizon_0_1000000_days")


def test_horizon_5_10_hours(tmp_path: pathlib.Path):
    # (5 hours, 10.5 hours)
    time_horizon = TimeHorizon(datetime.timedelta(hours=5), datetime.timedelta(hours=10, minutes=30))
    labeler = DummyLabeler([2], time_horizon)
    events_with_labels: EventsWithLabels = [
        # fmt: off
        (((2015, 1, 1, 0, 0), 1, None), True),
        (((2015, 1, 1, 10, 29), 2, None), False),
        (((2015, 1, 1, 10, 30), 1, None), False),
        (((2015, 1, 1, 10, 31), 1, None), False),
        #
        (((2016, 1, 1, 0, 0), 1, None), True),
        (((2016, 1, 1, 10, 29), 1, None), False),
        (((2016, 1, 1, 10, 30), 2, None), False),
        (((2016, 1, 1, 10, 31), 1, None), False),
        #
        (((2017, 1, 1, 0, 0), 1, None), False),
        (((2017, 1, 1, 10, 29), 1, None), False),
        (((2017, 1, 1, 10, 30), 1, None), False),
        (((2017, 1, 1, 10, 31), 2, None), False),
        #
        (((2018, 1, 1, 0, 0), 1, None), False),
        (((2018, 1, 1, 4, 59, 59), 2, None), False),
        (((2018, 1, 1, 5), 1, None), False),
        #
        (((2019, 1, 1, 0, 0), 1, None), True),
        (((2019, 1, 1, 4, 59, 59), 1, None), "out of range"),
        (((2019, 1, 1, 5), 2, None), "out of range"),
        # fmt: on
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_horizon_5_10_hours")


def test_horizon_infinite(tmp_path: pathlib.Path):
    # Infinite horizon
    time_horizon = TimeHorizon(
        datetime.timedelta(days=10),
        None,
    )
    labeler = DummyLabeler([2], time_horizon)
    events_with_labels: EventsWithLabels = [
        # fmt: off
        (((1950, 1, 3), 1, None), True),
        (((2000, 1, 3), 1, None), True),
        (((2001, 10, 5), 1, None), True),
        (((2020, 10, 5), 1, None), True),
        (((2021, 10, 5), 1, None), True),
        (((2050, 1, 10), 2, None), True),
        (((2050, 1, 20), 2, None), False),
        (((2051, 1, 10), 1, None), False),
        (((5000, 1, 10), 1, None), False),
        # fmt: on
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_horizon_infinite")



import datetime
from typing import Any, List, Mapping, cast

import femr_test_tools
import meds
import scipy.sparse

import femr
import femr.index
from femr.featurizers import FeaturizerList
from femr.featurizers.featurizers import AgeFeaturizer, CountFeaturizer
from femr.labelers import TimeHorizon
from femr.labelers.omop import CodeLabeler


def _assert_featurized_patients_structure(labels: List[meds.Label], features: Mapping[str, Any]):
    assert features["features"].dtype == "float32"
    assert features["patient_ids"].dtype == "int64"
    assert features["feature_times"].dtype == "datetime64[us]"

    assert features["feature_times"].shape[0] == len(labels)
    assert features["patient_ids"].shape[0] == len(labels)
    assert features["features"].shape[0] == len(labels)

    assert sorted(list(features["patient_ids"])) == sorted(list(label["patient_id"] for label in labels))
    assert sorted(list(features["feature_times"])) == sorted(list(label["prediction_time"] for label in labels))


def test_age_featurizer() -> None:
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))

    dataset = femr_test_tools.create_patients_dataset(100)
    index = femr.index.PatientIndex(dataset)

    labeler = CodeLabeler(["2"], time_horizon, ["3"])

    patient: meds.Patient = dataset[0]
    labels = labeler.label(patient)
    featurizer = AgeFeaturizer(is_normalize=False)
    patient_features = featurizer.featurize(patient, labels)

    assert patient_features[0] == [(0, 15.43013698630137)]
    assert patient_features[1] == [(0, 17.767123287671232)]
    assert patient_features[-1] == [(0, 20.46027397260274)]

    all_labels = labeler.apply(dataset)

    featurizer = AgeFeaturizer(is_normalize=True)
    featurizer_list = FeaturizerList([featurizer])
    featurizer_list.preprocess_featurizers(dataset, index, all_labels)
    featurized_patients = featurizer_list.featurize(dataset, index, all_labels)

    _assert_featurized_patients_structure(all_labels, featurized_patients)


def test_count_featurizer() -> None:
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))

    dataset = femr_test_tools.create_patients_dataset(100)
    index = femr.index.PatientIndex(dataset)

    labeler = CodeLabeler(["2"], time_horizon, ["3"])

    patient: meds.Patient = dataset[0]
    labels = labeler.label(patient)
    featurizer = CountFeaturizer()
    data = featurizer.generate_preprocess_data([patient], {patient["patient_id"]: labels})
    featurizer.encorperate_prepreprocessed_data([data])

    patient_features = featurizer.featurize(patient, labels)

    assert featurizer.get_num_columns() == 4, f"featurizer.get_num_columns() = {featurizer.get_num_columns()}"

    simple_patient_features = [{(featurizer.get_column_name(v.column), v.value) for v in a} for a in patient_features]

    assert simple_patient_features[0] == {
        ("SNOMED/184099003", 1),
        ("3", 1),
    }
    assert simple_patient_features[1] == {
        ("SNOMED/184099003", 1),
        ("3", 2),
        ("2", 2),
    }
    assert simple_patient_features[2] == {
        ("SNOMED/184099003", 1),
        ("3", 3),
        ("2", 4),
    }

    all_labels = labeler.apply(dataset)

    featurizer = CountFeaturizer()
    featurizer_list = FeaturizerList([featurizer])
    featurizer_list.preprocess_featurizers(dataset, index, all_labels)
    featurized_patients = featurizer_list.featurize(dataset, index, all_labels)

    _assert_featurized_patients_structure(all_labels, featurized_patients)


def test_count_featurizer_with_ontology() -> None:
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))

    dataset = femr_test_tools.create_patients_dataset(100)
    index = femr.index.PatientIndex(dataset)

    labeler = CodeLabeler(["2"], time_horizon, ["3"])

    patient: meds.Patient = dataset[0]
    labels = labeler.label(patient)

    class DummyOntology:
        def get_all_parents(self, code):
            if code in ("2", "SNOMED/184099003"):
                return {"parent", code}
            else:
                return {code}

    featurizer = CountFeaturizer(is_ontology_expansion=True, ontology=cast(femr.ontology.Ontology, DummyOntology()))
    data = featurizer.generate_preprocess_data([patient], {patient["patient_id"]: labels})
    featurizer.encorperate_prepreprocessed_data([data])

    patient_features = featurizer.featurize(patient, labels)

    assert featurizer.get_num_columns() == 5, f"featurizer.get_num_columns() = {featurizer.get_num_columns()}"

    simple_patient_features = [{(featurizer.get_column_name(v.column), v.value) for v in a} for a in patient_features]

    assert simple_patient_features[0] == {
        ("SNOMED/184099003", 1),
        ("3", 1),
        ("parent", 1),
    }
    assert simple_patient_features[1] == {
        ("SNOMED/184099003", 1),
        ("3", 2),
        ("2", 2),
        ("parent", 3),
    }
    assert simple_patient_features[2] == {
        ("SNOMED/184099003", 1),
        ("parent", 5),
        ("3", 3),
        ("2", 4),
    }

    all_labels = labeler.apply(dataset)

    featurizer = CountFeaturizer(is_ontology_expansion=True, ontology=cast(femr.ontology.Ontology, DummyOntology()))
    featurizer_list = FeaturizerList([featurizer])
    featurizer_list.preprocess_featurizers(dataset, index, all_labels)
    featurized_patients = featurizer_list.featurize(dataset, index, all_labels)

    _assert_featurized_patients_structure(all_labels, featurized_patients)


def test_count_featurizer_with_values() -> None:
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))

    dataset = femr_test_tools.create_patients_dataset(100)
    index = femr.index.PatientIndex(dataset)

    labeler = CodeLabeler(["2"], time_horizon, ["3"])

    patient: meds.Patient = dataset[0]
    labels = labeler.label(patient)
    featurizer = CountFeaturizer(numeric_value_decile=True, string_value_combination=True)
    data = featurizer.generate_preprocess_data([patient], {patient["patient_id"]: labels})
    featurizer.encorperate_prepreprocessed_data([data])

    patient_features = featurizer.featurize(patient, labels)

    assert featurizer.get_num_columns() == 7

    simple_patient_features = [{(featurizer.get_column_name(v.column), v.value) for v in a} for a in patient_features]

    assert simple_patient_features[0] == {
        ("SNOMED/184099003", 1),
        ("3", 1),
        ("2 [1.0, inf)", 1),
        ("1 test_value", 2),
    }

    assert simple_patient_features[1] == {
        ("SNOMED/184099003", 1),
        ("3", 2),
        ("2", 2),
        ("2 [1.0, inf)", 1),
        ("1 test_value", 2),
    }
    assert simple_patient_features[2] == {
        ("SNOMED/184099003", 1),
        ("3", 3),
        ("2", 4),
        ("2 [1.0, inf)", 1),
        ("1 test_value", 2),
    }

    all_labels = labeler.apply(dataset)

    featurizer = CountFeaturizer(numeric_value_decile=True, string_value_combination=True)
    featurizer_list = FeaturizerList([featurizer])
    featurizer_list.preprocess_featurizers(dataset, index, all_labels)
    featurized_patients = featurizer_list.featurize(dataset, index, all_labels)

    _assert_featurized_patients_structure(all_labels, featurized_patients)


def test_count_featurizer_exclude_filter() -> None:
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))

    dataset = femr_test_tools.create_patients_dataset(100)

    labeler = CodeLabeler(["2"], time_horizon, ["3"])

    patient: meds.Patient = dataset[0]
    labels = labeler.label(patient)

    # Test filtering all codes
    featurizer = CountFeaturizer(excluded_event_filter=lambda _: True)
    data = featurizer.generate_preprocess_data([patient], {patient["patient_id"]: labels})
    featurizer.encorperate_prepreprocessed_data([data])

    assert featurizer.get_num_columns() == 0

    # Test filtering no codes
    featurizer = CountFeaturizer(excluded_event_filter=lambda _: False)
    data = featurizer.generate_preprocess_data([patient], {patient["patient_id"]: labels})
    featurizer.encorperate_prepreprocessed_data([data])

    assert featurizer.get_num_columns() == 4

    # Test filtering single code
    featurizer = CountFeaturizer(excluded_event_filter=lambda e: e["code"] == "3")
    data = featurizer.generate_preprocess_data([patient], {patient["patient_id"]: labels})
    featurizer.encorperate_prepreprocessed_data([data])

    assert featurizer.get_num_columns() == 3


def test_count_bins_featurizer() -> None:
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))

    dataset = femr_test_tools.create_patients_dataset(100)
    index = femr.index.PatientIndex(dataset)

    labeler = CodeLabeler(["2"], time_horizon, ["3"])

    patient: meds.Patient = dataset[0]
    labels = labeler.label(patient)
    time_bins = [
        datetime.timedelta(days=90),
        datetime.timedelta(days=180),
        datetime.timedelta(weeks=1e4),
    ]
    featurizer = CountFeaturizer(
        time_bins=time_bins,
    )
    data = featurizer.generate_preprocess_data([patient], {patient["patient_id"]: labels})
    featurizer.encorperate_prepreprocessed_data([data])

    patient_features = featurizer.featurize(patient, labels)

    assert featurizer.get_num_columns() == 12

    simple_patient_features = [{(featurizer.get_column_name(v.column), v.value) for v in a} for a in patient_features]

    assert simple_patient_features[0] == {
        ("SNOMED/184099003_70000 days, 0:00:00", 1),
        ("3_90 days, 0:00:00", 1),
    }
    assert simple_patient_features[1] == {
        ("3_90 days, 0:00:00", 1),
        ("SNOMED/184099003_70000 days, 0:00:00", 1),
        ("3_70000 days, 0:00:00", 1),
        ("2_70000 days, 0:00:00", 2),
    }
    assert simple_patient_features[2] == {
        ("2_70000 days, 0:00:00", 2),
        ("2_90 days, 0:00:00", 2),
        ("SNOMED/184099003_70000 days, 0:00:00", 1),
        ("3_90 days, 0:00:00", 1),
        ("3_70000 days, 0:00:00", 2),
    }

    all_labels = labeler.apply(dataset)

    time_bins = [
        datetime.timedelta(days=90),
        datetime.timedelta(days=180),
        datetime.timedelta(weeks=1e4),
    ]
    featurizer = CountFeaturizer(
        time_bins=time_bins,
    )
    featurizer_list = FeaturizerList([featurizer])
    featurizer_list.preprocess_featurizers(dataset, index, all_labels)
    featurized_patients = featurizer_list.featurize(dataset, index, all_labels)

    _assert_featurized_patients_structure(all_labels, featurized_patients)


def test_complete_featurization() -> None:
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))

    dataset = femr_test_tools.create_patients_dataset(100)
    index = femr.index.PatientIndex(dataset)

    labeler = CodeLabeler(["2"], time_horizon, ["3"])

    all_labels = labeler.apply(dataset)

    age_featurizer = AgeFeaturizer(is_normalize=True)
    age_featurizer_list = FeaturizerList([age_featurizer])
    age_featurizer_list.preprocess_featurizers(dataset, index, all_labels)
    age_featurized_patients = age_featurizer_list.featurize(dataset, index, all_labels)

    time_bins = [
        datetime.timedelta(days=90),
        datetime.timedelta(days=180),
        datetime.timedelta(weeks=1e5),
    ]
    count_featurizer = CountFeaturizer(time_bins=time_bins)
    count_featurizer_list = FeaturizerList([count_featurizer])
    count_featurizer_list.preprocess_featurizers(dataset, index, all_labels)
    count_featurized_patients = count_featurizer_list.featurize(dataset, index, all_labels)

    age_featurizer = AgeFeaturizer(is_normalize=True)
    time_bins = [
        datetime.timedelta(days=90),
        datetime.timedelta(days=180),
        datetime.timedelta(weeks=1e5),
    ]
    count_featurizer = CountFeaturizer(time_bins=time_bins)
    featurizer_list = FeaturizerList([age_featurizer, count_featurizer])
    featurizer_list.preprocess_featurizers(dataset, index, all_labels)
    featurized_patients = featurizer_list.featurize(dataset, index, all_labels)

    assert featurized_patients["patient_ids"].shape == count_featurized_patients["patient_ids"].shape

    the_same = (
        featurized_patients["features"].toarray()
        == scipy.sparse.hstack((age_featurized_patients["features"], count_featurized_patients["features"])).toarray()
    )

    assert the_same.all()



"""A collection of general use transforms."""
import datetime
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

import meds


def remove_nones(
    patient: meds.Patient,
    do_not_apply_to_filter: Optional[Callable[[meds.Measurement], bool]] = None,
) -> meds.Patient:
    """Remove duplicate codes w/in same day if duplicate code has None value.

    There is no point having a NONE value in a timeline when we have an actual value within the same day.

    This removes those unnecessary NONE values.
    """
    do_not_apply_to_filter = do_not_apply_to_filter or (lambda _: False)
    has_value: Set[Tuple[str, datetime.date]] = set()

    for event in patient["events"]:
        for measurement in event["measurements"]:
            value = (measurement["numeric_value"], measurement["text_value"], measurement["datetime_value"])
            if any(v is not None for v in value):
                has_value.add((measurement["code"], event["time"].date()))

    new_events: List[meds.Event] = []
    for event in patient["events"]:
        new_measurements: List[meds.Measurement] = []

        for measurement in event["measurements"]:
            value = (measurement["numeric_value"], measurement["text_value"], measurement["datetime_value"])
            if (
                all(v is None for v in value)
                and (measurement["code"], event["time"].date()) in has_value
                and not do_not_apply_to_filter(measurement)
            ):
                # Skip this event as already in there
                continue

            new_measurements.append(measurement)

        if len(new_measurements) > 0:
            new_events.append({"time": event["time"], "measurements": new_measurements})

    patient["events"] = new_events
    patient["events"].sort(key=lambda a: a["time"])

    return patient


def delta_encode(
    patient: meds.Patient,
    do_not_apply_to_filter: Optional[Callable[[meds.Measurement], bool]] = None,
) -> meds.Patient:
    """Delta encodes the patient.

    The idea behind delta encoding is that if we get duplicate values within a short amount of time
    (1 day for this code), there is not much point retaining the duplicate.

    This code removes all *sequential* duplicates within the same day.
    """
    do_not_apply_to_filter = do_not_apply_to_filter or (lambda _: False)

    last_value: Dict[Tuple[str, datetime.date], Any] = {}

    new_events: List[meds.Event] = []
    for event in patient["events"]:
        new_measurements: List[meds.Measurement] = []
        for measurement in event["measurements"]:
            key = (measurement["code"], event["time"].date())
            value = (measurement["numeric_value"], measurement["text_value"], measurement["datetime_value"])
            if key in last_value and last_value[key] == value and not do_not_apply_to_filter(measurement):
                continue
            last_value[key] = value
            new_measurements.append(measurement)

        if len(new_measurements) > 0:
            new_events.append({"time": event["time"], "measurements": new_measurements})

    patient["events"] = new_events
    patient["events"].sort(key=lambda a: a["time"])

    return patient


def fix_events(patient: meds.Patient) -> meds.Patient:
    """After a series of transformations, sometimes the patient structure gets a bit messed up.
    The usual issues are either duplicate event times or missorted events.

    This does a final cleanup pass to meet the MEDS requirements.
    """
    patient["events"].sort(key=lambda a: a["time"])

    if len(patient["events"]) == 0:
        return patient

    new_events = []
    new_event: meds.Event = {"time": patient["events"][0]["time"], "measurements": []}
    for event in patient["events"]:
        if new_event["time"] != event["time"]:
            new_events.append(new_event)
            new_event = {"time": event["time"], "measurements": []}
        new_event["measurements"].extend(event["measurements"])
    new_events.append(new_event)
    patient["events"] = new_events

    return patient



import dataclasses
import math
import random


@dataclasses.dataclass
class OnlineStatistics:
    """
    A class for computing online statistics such as mean and variance.
    From https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance.
    """

    count: float
    current_mean: float
    variance: float

    def __init__(self):
        """
        Initialize online statistics. Optionally takes the results of self.to_dict() to initialize from old data.
        """
        self.count = 0
        self.current_mean = 0
        self.variance = 0

    def add(self, weight: float, value: float) -> None:
        """
        Add an observation to the calculation.
        """
        self.count += weight
        delta = value - self.current_mean
        self.current_mean += delta * (weight / self.count)
        delta2 = value - self.current_mean

        self.variance += weight * (delta * delta2)

    def mean(self) -> float:
        """
        Return the current mean.
        """
        return self.current_mean

    def standard_deviation(self) -> float:
        """
        Return the current standard devation.
        """
        return math.sqrt(self.variance / self.count)

    def combine(self, other) -> None:
        if self.count == 0:
            self.count = other.count
            self.current_mean = other.current_mean
            self.variance = other.variance
        elif other.count != 0:
            total = self.count + other.count
            delta = other.current_mean - self.current_mean
            new_mean = self.current_mean + delta * (other.count / total)
            new_variance = self.variance + other.variance + (delta * self.count) * (delta * other.count) / total

            self.count = total
            self.current_mean = new_mean
            self.variance = new_variance


class ReservoirSampler:
    def __init__(self, size):
        self.total_weight = 0
        self.size = size
        self.samples = []

    def add(self, sample, weight):
        self.total_weight += weight
        if len(self.samples) < self.size:
            self.samples.append(sample)
            if len(self.samples) == self.size:
                self.j = random.random()
                self.p_none = 1
        else:
            prob = weight / self.total_weight
            self.j -= prob * self.p_none
            self.p_none = self.p_none * (1 - prob)

            if self.j <= 0:
                self.samples[random.randint(0, self.size - 1)] = sample
                self.j = random.random()
                self.p_none = 1

    def combine(self, other):
        for val in other.samples:
            self.add(val, other.total_weight / len(other.samples))



"""meds.Labeling functions for OMOP data."""
from __future__ import annotations

import datetime
from typing import Any, Callable, List, Optional

import meds

import femr.ontology

from .core import TimeHorizon, TimeHorizonEventLabeler


def identity(x: Any) -> Any:
    return x


def get_death_concepts() -> List[str]:
    return [
        meds.death_code,
    ]


def move_datetime_to_end_of_day(date: datetime.datetime) -> datetime.datetime:
    return date.replace(hour=23, minute=59, second=0)


##########################################################
##########################################################
# Abstract classes derived from TimeHorizonEventLabeler
##########################################################
##########################################################


class CodeLabeler(TimeHorizonEventLabeler):
    """Apply a label based on 1+ outcome_codes' occurrence(s) over a fixed time horizon."""

    def __init__(
        self,
        outcome_codes: List[str],
        time_horizon: TimeHorizon,
        prediction_codes: Optional[List[str]] = None,
        prediction_time_adjustment_func: Callable = identity,
    ):
        """Create a CodeLabeler, which labels events whose index in your Ontology is in `self.outcome_codes`

        Args:
            prediction_codes (List[int]): Events that count as an occurrence of the outcome.
            time_horizon (TimeHorizon): An interval of time. If the event occurs during this time horizon, then
                the label is TRUE. Otherwise, FALSE.
            prediction_codes (Optional[List[int]]): If not None, limit events at which you make predictions to
                only events with an `event.code` in these codes.
            prediction_time_adjustment_func (Optional[Callable]). A function that takes in a `datetime.datetime`
                and returns a different `datetime.datetime`. Defaults to the identity function.
        """
        self.outcome_codes: List[str] = outcome_codes
        self.time_horizon: TimeHorizon = time_horizon
        self.prediction_codes: Optional[List[str]] = prediction_codes
        self.prediction_time_adjustment_func: Callable = prediction_time_adjustment_func

    def get_prediction_times(self, patient: meds.Patient) -> List[datetime.datetime]:
        """Return each event's start time (possibly modified by prediction_time_adjustment_func)
        as the time to make a prediction. Default to all events whose `code` is in `self.prediction_codes`."""
        times: List[datetime.datetime] = []
        last_time = None
        for e in patient["events"]:
            prediction_time: datetime.datetime = self.prediction_time_adjustment_func(e["time"])

            for m in e["measurements"]:
                if ((self.prediction_codes is None) or (m["code"] in self.prediction_codes)) and (
                    last_time != prediction_time
                ):
                    times.append(prediction_time)
                    last_time = prediction_time
        return times

    def get_time_horizon(self) -> TimeHorizon:
        return self.time_horizon

    def get_outcome_times(self, patient: meds.Patient) -> List[datetime.datetime]:
        """Return the start times of this patient's events whose `code` is in `self.outcome_codes`."""
        times: List[datetime.datetime] = []
        for event in patient["events"]:
            for measurement in event["measurements"]:
                if measurement["code"] in self.outcome_codes:
                    times.append(event["time"])
        return times

    def allow_same_time_labels(self) -> bool:
        # We cannot allow labels at the same time as the codes since they will generally be available as features ...
        return False


class OMOPConceptCodeLabeler(CodeLabeler):
    """Same as CodeLabeler, but add the extra step of mapping OMOP concept IDs
    (stored in `omop_concept_ids`) to femr codes (stored in `codes`)."""

    # parent OMOP concept codes, from which all the outcome
    # are derived (as children from our ontology)
    original_omop_concept_codes: List[str] = []

    def __init__(
        self,
        ontology: femr.ontology.Ontology,
        time_horizon: TimeHorizon,
        prediction_codes: Optional[List[str]] = None,
        prediction_time_adjustment_func: Callable = identity,
    ):
        outcome_codes: List[str] = []
        for code in self.original_omop_concept_codes:
            outcome_codes.extend(ontology.get_all_children(code))
        super().__init__(
            outcome_codes=outcome_codes,
            time_horizon=time_horizon,
            prediction_codes=prediction_codes,
            prediction_time_adjustment_func=prediction_time_adjustment_func,
        )


##########################################################
##########################################################
# meds.Labeling functions derived from CodeLabeler
##########################################################
##########################################################


class MortalityCodeLabeler(CodeLabeler):
    """Apply a label for whether or not a patient dies within the `time_horizon`.
    Make prediction at admission time.
    """

    def __init__(
        self,
        ontology: femr.ontology.Ontology,
        time_horizon: TimeHorizon,
        prediction_codes: Optional[List[str]] = None,
        prediction_time_adjustment_func: Callable = identity,
    ):
        """Create a Mortality labeler."""
        outcome_codes: List[str] = []
        for code in get_death_concepts():
            outcome_codes.extend(ontology.get_all_children(code))

        super().__init__(
            outcome_codes=outcome_codes,
            time_horizon=time_horizon,
            prediction_codes=prediction_codes,
            prediction_time_adjustment_func=prediction_time_adjustment_func,
        )


class LupusCodeLabeler(OMOPConceptCodeLabeler):
    """
    meds.Label if patient is diagnosed with Lupus.
    """

    original_omop_concept_codes = ["SNOMED/55464009", "SNOMED/201436003"]


##########################################################
##########################################################
# Labeling functions derived from OMOPConceptCodeLabeler
##########################################################
##########################################################


class HypoglycemiaCodeLabeler(OMOPConceptCodeLabeler):
    """Apply a label for whether a patient has at 1+ explicitly
    coded occurrence(s) of Hypoglycemia in `time_horizon`."""

    # fmt: off
    original_omop_concept_codes = [
        'SNOMED/267384006', 'SNOMED/421725003', 'SNOMED/719216001',
        'SNOMED/302866003', 'SNOMED/237633009', 'SNOMED/120731000119103',
        'SNOMED/190448007', 'SNOMED/230796005', 'SNOMED/421437000',
        'SNOMED/52767006', 'SNOMED/237637005', 'SNOMED/84371000119108'
    ]
    # fmt: on


class AKICodeLabeler(OMOPConceptCodeLabeler):
    """Apply a label for whether a patient has at 1+ explicitly
    coded occurrence(s) of AKI in `time_horizon`."""

    # fmt: off
    original_omop_concept_codes = [
        'SNOMED/14669001', 'SNOMED/298015003', 'SNOMED/35455006',
    ]
    # fmt: on


class AnemiaCodeLabeler(OMOPConceptCodeLabeler):
    """Apply a label for whether a patient has at 1+ explicitly
    coded occurrence(s) of Anemia in `time_horizon`."""

    # fmt: off
    original_omop_concept_codes = [
        'SNOMED/271737000', 'SNOMED/713496008', 'SNOMED/713349004', 'SNOMED/767657005',
        'SNOMED/111570005', 'SNOMED/691401000119104', 'SNOMED/691411000119101',
    ]
    # fmt: on


class HyperkalemiaCodeLabeler(OMOPConceptCodeLabeler):
    """Apply a label for whether a patient has at 1+ explicitly
    coded occurrence(s) of Hyperkalemia in `time_horizon`."""

    # fmt: off
    original_omop_concept_codes = [
        'SNOMED/14140009',
    ]
    # fmt: on


class HyponatremiaCodeLabeler(OMOPConceptCodeLabeler):
    """Apply a label for whether a patient has at 1+ explicitly
    coded occurrence(s) of Hyponatremia in `time_horizon`."""

    # fmt: off
    original_omop_concept_codes = [
        'SNOMED/267447008', 'SNOMED/89627008'
    ]
    # fmt: on


class ThrombocytopeniaCodeLabeler(OMOPConceptCodeLabeler):
    """Apply a label for whether a patient has at 1+ explicitly
    coded occurrence(s) of Thrombocytopenia in `time_horizon`."""

    # fmt: off
    original_omop_concept_codes = [
        'SNOMED/267447008', 'SNOMED/89627008',
    ]
    # fmt: on


class NeutropeniaCodeLabeler(OMOPConceptCodeLabeler):
    """Apply a label for whether a patient has at 1+ explicitly
    coded occurrence(s) of Neutkropenia in `time_horizon`."""

    # fmt: off
    original_omop_concept_codes = [
        'SNOMED/165517008',
    ]
    # fmt: on



import numpy as np
import pytest

from femr.featurizers.utils import OnlineStatistics


def _assert_correct_stats(stat: OnlineStatistics, values: list):
    TOLERANCE = 1e-6  # Allow for some floating point error
    true_mean = np.mean(values)
    true_sample_variance = np.var(values, ddof=1)
    true_m2 = true_sample_variance * (len(values) - 1)
    assert stat.current_count == len(values), f"{stat.current_count} != {len(values)}"
    assert np.isclose(stat.mean(), true_mean), f"{stat.mean()} != {true_mean}"
    assert np.isclose(
        stat.variance(), true_sample_variance, atol=TOLERANCE
    ), f"{stat.variance()} != {true_sample_variance}"
    assert np.isclose(stat.current_M2, true_m2, atol=TOLERANCE), f"{stat.current_M2} != {true_m2}"


def test_add():
    # Test adding things to the statistics
    def _run_test(values):
        stat = OnlineStatistics()
        for i in values:
            stat.add(i)
        _assert_correct_stats(stat, values)

    # Positive integers
    _run_test(range(51))
    _run_test(range(10, 10000, 3))
    # Negative integers
    _run_test(range(-400, -300))
    # Positive/negative integers
    _run_test(list(range(4, 900, 2)) + list(range(-1000, -300, 7)))
    _run_test(list(range(-100, 100, 7)) + list(range(-100, 100, 2)))
    # Decimals
    _run_test(np.linspace(0, 1, 100))
    _run_test(np.logspace(-100, 3, 100))
    # Small lists
    _run_test([0, 1])
    _run_test([-1, 1])


def test_constructor():
    # Test default
    stat = OnlineStatistics()
    assert stat.current_count == 0
    assert stat.current_mean == stat.mean() == 0
    assert stat.current_M2 == 0

    # Test explicitly setting args
    stat = OnlineStatistics(current_count=1, current_mean=2, current_variance=3)
    assert stat.current_count == 1
    assert stat.current_mean == stat.mean() == 2
    assert stat.current_M2 == 0

    # Test M2
    stat = OnlineStatistics(current_count=10, current_mean=20, current_variance=30)
    assert stat.current_count == 10
    assert stat.current_mean == 20
    assert stat.current_M2 == 30 * (10 - 1)

    # Test getters/setters
    stat = OnlineStatistics(current_count=10, current_mean=20, current_variance=30)
    assert stat.mean() == 20
    assert stat.variance() == 30
    assert stat.standard_deviation() == np.sqrt(30)

    # Test fail cases
    with pytest.raises(ValueError) as _:
        # Negative count
        stat = OnlineStatistics(current_count=-1, current_mean=2, current_variance=3)
    with pytest.raises(ValueError) as _:
        # Negative variance
        stat = OnlineStatistics(current_count=1, current_mean=2, current_variance=-3)
    with pytest.raises(ValueError) as _:
        # Positive variance with 0 count
        stat = OnlineStatistics(current_count=0, current_mean=2, current_variance=1)
    with pytest.raises(ValueError) as _:
        # Can only compute variance with >1 observation
        stat = OnlineStatistics()
        stat.add(1)
        stat.variance()


def test_merge_pair():
    # Simulate two statistics being merged via `merge_pair``
    stat1 = OnlineStatistics()
    values1 = list(range(-300, 300, 4)) + list(range(400, 450))
    for i in values1:
        stat1.add(i)
    stat2 = OnlineStatistics()
    values2 = list(range(100, 150))
    for i in values2:
        stat2.add(i)
    merged_stat = OnlineStatistics.merge_pair(stat1, stat2)
    merged_stat_values = values1 + values2
    _assert_correct_stats(merged_stat, merged_stat_values)


def test_merge():
    # Simulate parallel statistics being merged via `merge`
    stats = []
    values = [
        np.linspace(-100, 100, 50),
        np.linspace(100, 200, 50),
        np.linspace(100, 150, 100),
        np.linspace(-10, 100, 100),
        np.linspace(10, 200, 3),
    ]
    for i in range(len(values)):
        stat = OnlineStatistics()
        for v in values[i]:
            stat.add(v)
        stats.append(stat)
    merged_stat = OnlineStatistics.merge(stats)
    _assert_correct_stats(merged_stat, np.concatenate(values))



from __future__ import annotations

import copy
import math
from typing import List


class OnlineStatistics:
    """
    A class for computing online statistics such as mean and variance.
    Uses Welford's online algorithm.
    From https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm.

    NOTE: The variance we calculate is the sample variance, not the population variance.
    """

    def __init__(
        self,
        current_count: int = 0,
        current_mean: float = 0,
        current_variance: float = 0,
    ):
        """
        Initialize online statistics.
            `mean` accumulates the mean of the entire dataset
            `count` aggregates the number of samples seen so far
            `current_M2` aggregates the squared distances from the mean
        """
        if not (current_count >= 0 and current_variance >= 0):
            raise ValueError(
                "Must set `current_count` and `current_variance` to be non-negative."
                f"You specified `current_count` = {current_count} and `current_variance` = {current_variance}."
            )
        self.current_count: int = current_count
        self.current_mean: float = current_mean
        if current_count == 0 and current_variance == 0:
            self.current_M2 = 0.0
        elif current_count > 0:
            self.current_M2 = current_variance * (current_count - 1)
        else:
            raise ValueError(
                "Cannot specify `current_variance` with a value > 0"
                "without specifying `current_count` with a value > 0. "
                f"You specified `current_count` = {current_count} and `current_variance` = {current_variance}."
            )

    def add(self, newValue: float) -> None:
        """
        Add an observation to the calculation using Welford's online algorithm.

        Taken from: https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm
        """
        self.current_count += 1
        delta: float = newValue - self.current_mean
        self.current_mean += delta / self.current_count
        delta2: float = newValue - self.current_mean
        self.current_M2 += delta * delta2

    def mean(self) -> float:
        """
        Return the current mean.
        """
        return self.current_mean

    def variance(self) -> float:
        """
        Return the current sample variance.
        """
        if self.current_count < 2:
            raise ValueError(f"Cannot compute variance with only {self.current_count} observations.")

        return self.current_M2 / (self.current_count - 1)

    def standard_deviation(self) -> float:
        """
        Return the current standard devation.
        """
        return math.sqrt(self.variance())

    @classmethod
    def merge_pair(cls, stats1: OnlineStatistics, stats2: OnlineStatistics) -> OnlineStatistics:
        """
        Merge two sets of online statistics using Chan's parallel algorithm.

        Taken from: https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
        """
        if stats1.current_count == 0:
            return stats2
        elif stats2.current_count == 0:
            return stats1

        count: int = stats1.current_count + stats2.current_count
        delta: float = stats2.current_mean - stats1.current_mean
        mean: float = stats1.current_mean + delta * stats2.current_count / count
        M2 = stats1.current_M2 + stats2.current_M2 + delta**2 * stats1.current_count * stats2.current_count / count
        return OnlineStatistics(count, mean, M2 / (count - 1))

    @classmethod
    def merge(cls, stats_list: List[OnlineStatistics]) -> OnlineStatistics:
        """
        Merge a list of online statistics.
        """
        if len(stats_list) == 0:
            raise ValueError("Cannot merge an empty list of statistics.")
        unmerged_stats: List[OnlineStatistics] = copy.deepcopy(stats_list)
        # Run tree reduction to merge together all pairs of statistics
        # in a numerically stable way
        #   Example: 1 2 3 4 5 -> 3 7 5 -> 10 5 -> 15
        while len(unmerged_stats) > 1:
            merged_stats: List[OnlineStatistics] = []
            for i in range(0, len(unmerged_stats), 2):
                if i + 1 < len(unmerged_stats):
                    # If there's another stat after this one, merge them
                    merged_stats.append(cls.merge_pair(unmerged_stats[i], unmerged_stats[i + 1]))
                else:
                    # We've reached the end of our list, so just add the last stat back
                    merged_stats.append(unmerged_stats[i])
            unmerged_stats = merged_stats
        assert len(unmerged_stats) == 1, f"Should only have one stat left after merging, not ({len(unmerged_stats)})."
        return unmerged_stats[0]



from __future__ import annotations

import collections
import datetime
import math
from typing import Any, Dict, List, Mapping, Optional, Tuple

import datasets
import meds
import numpy as np
import torch
import torch.nn.functional as F
import transformers
import xformers.ops
from torch import nn

import femr.models.config
import femr.models.processor
import femr.models.rmsnorm
import femr.models.tasks
import femr.models.tokenizer
import femr.models.xformers


# From https://github.com/kingoflolz/mesh-transformer-jax
def rotate_every_two_v2(x):
    flat_x = x.reshape(-1, x.shape[-1])

    x1 = flat_x[:, ::2]
    x2 = flat_x[:, 1::2]

    result = torch.stack((-x2, x1), axis=-1).reshape(x.shape)

    assert x.dtype == result.dtype
    return result


def fixed_pos_embedding(ages, dim, dtype):
    assert ages.dtype == torch.float32
    assert len(ages.shape) == 1

    inv_freq = 1.0 / (10000 ** (torch.linspace(0, 2, steps=dim // 2, device=ages.device)))
    inv_freq = inv_freq.reshape(1, 1, dim // 2)
    assert inv_freq.dtype == torch.float32

    ages = ages.reshape(ages.shape[0], 1)

    t = inv_freq * ages

    sin, cos = torch.sin(t), torch.cos(t)

    final_shape = (ages.shape[0], 1, dim)

    sin = torch.stack((sin, sin), axis=-1).reshape(final_shape).type(dtype)
    cos = torch.stack((cos, cos), axis=-1).reshape(final_shape).type(dtype)

    return sin, cos


def apply_rotary_pos_emb(x, sincos):
    sin, cos = sincos
    sin = sin.to(dtype=x.dtype)
    cos = cos.to(dtype=x.dtype)

    assert x.dtype == sin.dtype == cos.dtype, f"{x.dtype} {sin.dtype} {cos.dtype}"

    if len(sin.shape) != len(x.shape):
        new_shape = (1,) + sin.shape
        sin = sin.reshape(new_shape)
        cos = cos.reshape(new_shape)

    return (x * cos) + (rotate_every_two_v2(x) * sin)


class FEMREncoderLayer(nn.Module):
    def __init__(self, config: femr.models.config.FEMRTransformerConfig):
        super().__init__()
        self.config = config
        self.norm = femr.models.rmsnorm.RMSNorm(self.config.hidden_size)
        if self.config.hidden_act == "swiglu":
            hidden_mult = 2
        else:
            hidden_mult = 1

        self.input_proj = nn.Linear(
            self.config.hidden_size,
            self.config.hidden_size * 3 + hidden_mult * self.config.intermediate_size,
            bias=self.config.use_bias,
        )
        self.output_proj = nn.Linear(
            self.config.hidden_size + self.config.intermediate_size, self.config.hidden_size, bias=self.config.use_bias
        )

    def forward(self, x, normed_ages, pos_embed, attn_bias):
        x = self.norm(x)

        if self.config.use_normed_ages:
            x[:, -2] = normed_ages.to(dtype=x.dtype)
            x[:, -1] = (normed_ages**2).to(dtype=x.dtype)

        transformed = self.input_proj(x)

        ff = transformed[:, : -self.config.hidden_size * 3]
        qkv = transformed[:, -self.config.hidden_size * 3 :]

        head_size = self.config.hidden_size // self.config.n_heads

        qkv = qkv.reshape(x.shape[0], 3, self.config.n_heads, head_size)

        q = apply_rotary_pos_emb(qkv[:, 0, :, :], pos_embed)
        k = apply_rotary_pos_emb(qkv[:, 1, :, :], pos_embed)
        v = qkv[:, 2, :, :]

        attn = femr.models.xformers.memory_efficient_attention_wrapper(
            q.unsqueeze(0),
            k.unsqueeze(0),
            v.unsqueeze(0),
            attn_bias=attn_bias,
        )

        attn = attn.reshape(x.shape)

        if self.config.hidden_act == "gelu":
            ff = F.gelu(ff)
        elif self.config.hidden_act == "swiglu":
            x1, x2 = ff.chunk(2, dim=-1)
            ff = F.silu(x1) * x2

        combined = torch.concatenate((attn, ff), axis=-1)
        result = self.output_proj(combined)

        return result


class FEMRTransformer(nn.Module):
    def __init__(self, config: femr.models.config.FEMRTransformerConfig):
        super().__init__()
        self.config = config

        self.in_norm = femr.models.rmsnorm.RMSNorm(self.config.hidden_size)
        self.out_norm = femr.models.rmsnorm.RMSNorm(self.config.hidden_size)

        if not self.config.is_hierarchical:
            self.embed = nn.Embedding(self.config.vocab_size, self.config.hidden_size)
        else:
            self.embed_bag = nn.EmbeddingBag(
                num_embeddings=self.config.vocab_size,
                embedding_dim=self.config.hidden_size,
                mode="sum",
                include_last_offset=True,
            )

        self.layers = nn.ModuleList([FEMREncoderLayer(config) for _ in range(self.config.n_layers)])

    def forward(self, batch):
        if not self.config.is_hierarchical:
            x = self.embed(batch["tokens"])
        else:
            x = self.embed_bag(batch["hierarchical_tokens"], batch["token_indices"], batch["hierarchical_weights"])

        x = self.in_norm(x)
        normed_ages = batch["normalized_ages"]
        pos_embed = fixed_pos_embedding(batch["ages"], self.config.hidden_size // self.config.n_heads, x.dtype)

        attn_bias = xformers.ops.fmha.attn_bias.BlockDiagonalMask.from_seqlens(
            batch["patient_lengths"].tolist()
        ).make_local_attention(self.config.attention_width)

        for layer in self.layers:
            x = x + layer(x, normed_ages, pos_embed, attn_bias)

        final = self.out_norm(x)

        return final


class LabeledPatientTaskHead(nn.Module):
    def __init__(self, hidden_size: int):
        super().__init__()

    def forward(self, features: torch.Tensor, batch: Mapping[str, torch.Tensor], return_logits=False):
        return 0, {}


class CLMBRTaskHead(nn.Module):
    def __init__(self, hidden_size: int, clmbr_vocab_size: int):
        super().__init__()

        self.final_layer = nn.Linear(hidden_size, clmbr_vocab_size)

    def forward(self, features: torch.Tensor, batch: Mapping[str, torch.Tensor], return_logits=False):
        logits = self.final_layer(features)
        labels = batch["labels"]
        loss = F.cross_entropy(logits, labels)

        if not return_logits:
            logits = None

        return loss, {'logits': logits}


class MOTORTaskHead(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        pretraining_task_info: List[Tuple[str, float]],
        time_bins: List[float],
        final_layer_size: int,
    ):
        super().__init__()

        self.num_time_bins = len(time_bins) - 1
        self.num_tasks = len(pretraining_task_info)

        self.final_layer_size = final_layer_size
        self.final_layer = nn.Linear(hidden_size, self.num_time_bins * final_layer_size)

        self.task_layer = nn.Linear(self.final_layer_size, self.num_tasks)
        start_bias = torch.log2(torch.tensor([a[1] for a in pretraining_task_info], dtype=torch.float32))
        self.task_layer.bias.data = start_bias

    def forward(self, features: torch.Tensor, batch: Mapping[str, torch.Tensor], return_logits=False):
        time_independent_features = self.final_layer(features).reshape(
            features.shape[0], self.num_time_bins, self.final_layer_size
        )

        time_dependent_logits = self.task_layer(time_independent_features)

        assert (
            batch["log_time"].shape == time_dependent_logits.shape
        ), f"{time_dependent_logits.shape} {batch['log_time'].shape}"
        assert (
            batch["is_event"].shape == time_dependent_logits.shape
        ), f"{time_dependent_logits.shape} {batch['is_event'].shape}"

        survival_loss = torch.exp2(time_dependent_logits + batch["log_time"]).mean()
        event_loss = -math.log(2) * torch.where(batch["is_event"], time_dependent_logits, 0).mean()

        loss = survival_loss + event_loss

        if not return_logits:
            time_dependent_logits = None

        return loss, {'time_dependent_logits': time_dependent_logits}


def remove_first_dimension(data: Any) -> Any:
    if isinstance(data, collections.abc.Mapping):
        return {k: remove_first_dimension(v) for k, v in data.items()}
    elif isinstance(data, torch.Tensor):
        assert data.shape[0] == 1
        return data.squeeze(dim=0)
    elif isinstance(data, np.ndarray):
        assert data.shape[0] == 1
        return np.squeeze(data, axis=0)
    elif isinstance(data, (int, float, np.number, np.bool_)):
        return data
    else:
        raise RuntimeError("Could not convert item of type " + str(type(data)))


class FEMRModel(transformers.PreTrainedModel):
    config_class = femr.models.config.FEMRModelConfig

    def __init__(self, config: femr.models.config.FEMRModelConfig, **kwargs):
        # Allow the task config to be ovewritten
        if "task_config" in kwargs:
            config.task_config = kwargs["task_config"]

        super().__init__(config)

        self.transformer = FEMRTransformer(self.config.transformer_config)
        if self.config.task_config is not None:
            self.task_model = self.create_task_head()

    def create_task_head(self) -> nn.Module:
        hidden_size = self.config.transformer_config.hidden_size
        task_type = self.config.task_config.task_type
        task_kwargs = self.config.task_config.task_kwargs
        if task_type == "clmbr":
            return CLMBRTaskHead(hidden_size, **task_kwargs)
        elif task_type == "labeled_patients":
            return LabeledPatientTaskHead(hidden_size, **task_kwargs)
        elif task_type == "motor":
            return MOTORTaskHead(hidden_size, **task_kwargs)

    def forward(self, batch: Mapping[str, Any], return_loss=True, return_logits=False, return_reprs=False):
        # Need a return_loss parameter for transformers.Trainer to work properly
        assert return_loss

        batch = remove_first_dimension(batch)

        features = self.transformer(batch["transformer"])
        if "task" in batch and self.config.task_config is not None:
            features = features.reshape(-1, features.shape[-1])
            features = features[batch["transformer"]["label_indices"], :]
            loss, result = self.task_model(features, batch["task"], return_logits=return_logits)
            if return_reprs:
                result["representations"] = features
            if return_logits or return_reprs:
                result["timestamps"] = batch["transformer"]["timestamps"][batch["transformer"]["label_indices"]]
                result["patient_ids"] = batch["patient_ids"][batch["transformer"]["label_indices"]]
            return loss, result
        else:
            loss = 0
            features = features.reshape(-1, features.shape[-1])
            if 'task' in batch:
                features = features[batch["transformer"]["label_indices"], :]
                result = {
                    "timestamps": batch["transformer"]["timestamps"][batch["transformer"]["label_indices"]],
                    "patient_ids": batch["patient_ids"][batch["transformer"]["label_indices"]],
                    "representations": features,
                }
            else:
                result = {
                    "timestamps": batch["transformer"]["timestamps"],
                    "patient_ids": batch["patient_ids"],
                    "representations": features,
                }

            return loss, result



def compute_features(
    dataset: datasets.Dataset,
    model_path: str,
    labels: List[meds.Label],
    num_proc: int = 1,
    tokens_per_batch: int = 1024,
    device: Optional[torch.device] = None,
    ontology: Optional[femr.ontology.Ontology] = None,
) -> Dict[str, np.ndarray]:
    task = femr.models.tasks.LabeledPatientTask(labels)

    index = femr.index.PatientIndex(dataset, num_proc=num_proc)

    model = femr.models.transformer.FEMRModel.from_pretrained(model_path, task_config=task.get_task_config())
    tokenizer = femr.models.tokenizer.FEMRTokenizer.from_pretrained(model_path, ontology=ontology)
    processor = femr.models.processor.FEMRBatchProcessor(tokenizer, task=task)

    filtered_data = task.filter_dataset(dataset, index)

    if device:
        model = model.to(device)

    batches = processor.convert_dataset(
        filtered_data, tokens_per_batch=tokens_per_batch, min_patients_per_batch=1, num_proc=num_proc
    )

    batches.set_format("pt", device=device)

    all_patient_ids = []
    all_feature_times = []
    all_representations = []

    for batch in batches:
        batch = processor.collate([batch])["batch"]
        with torch.no_grad():
            _, result = model(batch, return_reprs=True)
            all_patient_ids.append(result["patient_ids"].cpu().numpy())
            all_feature_times.append(result["timestamps"].cpu().numpy())
            all_representations.append(result["representations"].cpu().numpy())

    return {
        "patient_ids": np.concatenate(all_patient_ids),
        "feature_times": np.concatenate(all_feature_times).astype('datetime64[s]'),
        "features": np.concatenate(all_representations),
    }



# flake8: noqa: E501

import pathlib

import femr.ontology


def create_fake_athena(tmp_path: pathlib.Path) -> pathlib.Path:
    athena = tmp_path / "athena"
    athena.mkdir()

    concept = athena / "CONCEPT.csv"
    concept.write_text(
        """concept_id	concept_name	domain_id	vocabulary_id	concept_class_id	standard_concept	concept_code	valid_start_date	valid_end_date	invalid_reason
37200198\tType 2 diabetes mellitus with mild nonproliferative diabetic retinopathy with macular edema, right eye\tCondition\tICD10CM\t7-char billing code\t\tE11.3211\t19700101\t20991231\t
380097\tMacular edema due to diabetes mellitus\tCondition\tSNOMED\tClinical Finding\tS\t312912001\t20020131\t20991231
4334884	Disorder of macula due to diabetes mellitus	Condition	SNOMED	Clinical Finding	S	232020009	20020131	20991231
4174977	Retinopathy due to diabetes mellitus	Condition	SNOMED	Clinical Finding	S	4855003	20020131	20991231
4208223	Disorder of macula of retina	Condition	SNOMED	Clinical Finding	S	312999006	20020131	20991231
4290333	Macular retinal edema	Condition	SNOMED	Clinical Finding	S	37231002	20020131	20991231
35626904	Retinal edema due to diabetes mellitus	Condition	SNOMED	Clinical Finding	S	770323005	20180731	20991231
45757435	Mild nonproliferative retinopathy due to type 2 diabetes mellitus	Condition	SNOMED	Clinical Finding	S	138911000119106	20150131	20991231
"""
    )
    relationship = athena / "CONCEPT_RELATIONSHIP.csv"
    relationship.write_text(
        """concept_id_1	concept_id_2	relationship_id	valid_start_date	valid_end_date	invalid_reason
37200198	380097	Maps to	20171001	20991231
37200198	1567956	Is a	20170428	20991231
37200198	1567959	Is a	20170428	20991231
37200198	45757435	Maps to	20171001	20991231
37200198	1567961	Is a	20170428	20991231
37200198	45552385	Is a	20170428	20991231
35977781	35977781	Mapped from	20200913	20991231
46135811	40642538	Has status	20220128	20991231
46135811	35631990	Has Module	20220128	20991231"""
    )

    ancestor = athena / "CONCEPT_ANCESTOR.csv"
    ancestor.write_text(
        """ancestor_concept_id\tdescendant_concept_id\tmin_levels_of_separation\tmax_levels_of_separation
373499	4334884	4	6
442793	4334884	3	3
255919	4334884	6	9
433128	4334884	4	4
4180628	4334884	6	8
4209989	4334884	3	3
441840	4334884	6	10
4274025	4334884	5	9
4082284	4334884	2	2
4042836	4334884	5	7
443767	4334884	2	2
4038502	4334884	5	8
4174977	4334884	1	1
4334884	4334884	0	0
4134440	4334884	5	7
4247371	4334884	5	8
375252	4334884	3	5
4027883	4334884	4	4
4208223	4334884	1	1
378416	4334884	2	2
4080992	4334884	4	6
4162092	4334884	3	3
255919	380097	7	10
433128	380097	5	5
442793	380097	4	4
373499	380097	5	7
4180628	380097	7	9
4209989	380097	4	4
37018677	380097	3	3
441840	380097	5	11
380097	380097	0	0
4274025	380097	4	10
372903	380097	2	2
433595	380097	4	4
4042836	380097	6	8
4082284	380097	3	3
443767	380097	3	3
4038502	380097	6	9
4174977	380097	2	2
4334884	380097	1	1
4134440	380097	6	8
4247371	380097	6	9
4290333	380097	1	1
375252	380097	4	6
4027883	380097	5	5
4040388	380097	3	3
4208223	380097	2	2
35626904	380097	1	1
378416	380097	3	3
4080992	380097	5	7
4162092	380097	4	4
"""
    )

    return athena


def test_only_athena(tmp_path: pathlib.Path) -> None:
    fake_athena = create_fake_athena(tmp_path)

    ontology = femr.ontology.Ontology(str(fake_athena))

    assert (
        ontology.get_description("ICD10CM/E11.3211")
        == "Type 2 diabetes mellitus with mild nonproliferative diabetic retinopathy with macular edema, right eye"
    )

    assert ontology.get_parents("ICD10CM/E11.3211") == {"SNOMED/312912001", "SNOMED/138911000119106"}
    assert ontology.get_parents("SNOMED/312912001") == {"SNOMED/37231002", "SNOMED/232020009", "SNOMED/770323005"}

    assert ontology.get_all_parents("ICD10CM/E11.3211") == {
        "SNOMED/37231002",
        "SNOMED/138911000119106",
        "SNOMED/4855003",
        "SNOMED/312999006",
        "SNOMED/312912001",
        "ICD10CM/E11.3211",
        "SNOMED/770323005",
        "SNOMED/232020009",
    }

    assert ontology.get_children("SNOMED/312912001") == {"ICD10CM/E11.3211"}
    assert ontology.get_children("SNOMED/37231002") == {"SNOMED/312912001"}

    assert ontology.get_all_children("SNOMED/37231002") == {"ICD10CM/E11.3211", "SNOMED/312912001", "SNOMED/37231002"}


def test_athena_and_custom(tmp_path: pathlib.Path) -> None:
    fake_athena = create_fake_athena(tmp_path)

    code_metadata = {
        "CUSTOM/CustomDiabetes": {"description": "A nice diabetes code", "parent_codes": ["ICD10CM/E11.3211"]}
    }

    ontology = femr.ontology.Ontology(str(fake_athena), code_metadata)

    assert ontology.get_description("CUSTOM/CustomDiabetes") == "A nice diabetes code"
    assert ontology.get_all_parents("CUSTOM/CustomDiabetes") == {
        "CUSTOM/CustomDiabetes",
        "SNOMED/37231002",
        "SNOMED/138911000119106",
        "SNOMED/4855003",
        "SNOMED/312999006",
        "SNOMED/312912001",
        "ICD10CM/E11.3211",
        "SNOMED/770323005",
        "SNOMED/232020009",
    }

    assert ontology.get_all_children("SNOMED/37231002") == {
        "CUSTOM/CustomDiabetes",
        "ICD10CM/E11.3211",
        "SNOMED/312912001",
        "SNOMED/37231002",
    }
